import logging

logger = logging.getLogger(__name__)


from pydantic_universal_settings.main import (
    add_settings as add_settings,
    BaseSettings as BaseSettings,
    EnvFileMixin as EnvFileMixin,
    CLIMixin as CLIMixin,
    generate_all_settings as generate_all_settings,
    init_settings as init_settings,
    get_settings_proxy as get_settings_proxy,
)



import abc
import json
import os
import tempfile
from pathlib import Path
from typing import Any, Dict, Generic, Iterable, List, Optional, Type, TypeVar, Union

from psutil import pid_exists
from pydantic import BaseModel, BaseSettings as PydanticBaseSettings, root_validator
from pydantic_universal_settings import logger

_global_settings_models_dict: Dict[str, Type[BaseModel]] = {}
_global_settings_models_names: List[str] = []


def add_settings(name: Optional[Union[str, Type[BaseModel]]] = None) -> Any:
    def decorator(cls: Type[BaseModel]) -> Type[BaseModel]:
        if isinstance(name, str) and name:
            model_name = name
        else:
            model_name = cls.__name__
        if model_name in _global_settings_models_dict:
            raise KeyError(f"Error: settings {model_name} already added!")
        # else:
        #     print(model_name)
        _global_settings_models_dict[model_name] = cls
        _global_settings_models_names.append(model_name)
        return cls

    if name is None or isinstance(name, str):
        return decorator
    else:
        return decorator(name)


def parse_settings_models(
    settings_models: Union[None, str, Iterable[str]]
) -> List[Type[BaseModel]]:
    _settings_models = []
    if settings_models is None:
        _settings_models = _global_settings_models_names
    elif isinstance(settings_models, str):
        _settings_models = [settings_models]
    elif isinstance(settings_models, Iterable):
        _settings_models = list(settings_models)
    result = []
    for model_name in _settings_models:
        if model_name not in _global_settings_models_dict:
            raise KeyError(f"Error: settings {model_name} not added!")
        result.append(_global_settings_models_dict[model_name])
    return result


class BaseSettings(BaseModel, abc.ABC):
    pass


T = TypeVar("T", bound=BaseSettings)


class EnvFileMixin(BaseSettings):
    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"


class CLIMixin(BaseModel):
    @staticmethod
    def _save(cli_settings: Dict[str, Any]) -> None:
        return

    @staticmethod
    def _load() -> Dict[str, Any]:
        return {}

    @root_validator(allow_reuse=True)
    def _inject_cli(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        from pydantic_universal_settings.cli import cli_settings

        if "__not_empty" in cli_settings:
            del cli_settings["__not_empty"]
            _cli_settings = cli_settings
            cls._save(cli_settings)
        else:
            _cli_settings = cls._load()
        for key, value in _cli_settings.items():
            if key in values and not key.startswith("__") and value is not None:
                values[key] = value
        return values


class CLIWatchMixin(CLIMixin):
    _temp_settings_file: Optional[str] = None

    @staticmethod
    def _save(cli_settings: Dict[str, Any]) -> None:
        if len(cli_settings) == 0:
            return
        tempdir = Path(tempfile.gettempdir()) / "fastapi_rest_framework"
        tempdir.mkdir(exist_ok=True)
        # delete previous settings files if the process doesn't exist
        for file in tempdir.glob("*.settings.json"):
            try:
                pid = int(file.name.split(".")[0])
                if not pid_exists(pid):
                    file.unlink(missing_ok=True)
            except:  # noqa
                pass
        file_path = tempdir / f"{os.getpid()}.settings.json"
        with open(file_path, "w") as fp:
            json.dump(cli_settings, fp)
        logger.info("Save command line options into {}.", file_path)

    @staticmethod
    def _load() -> Dict[str, Any]:
        tempdir = Path(tempfile.gettempdir()) / "pydantic_universal_settings"
        file_path = tempdir / f"{os.getppid()}.cli.json"
        try:
            with open(file_path) as fp:
                data = json.load(fp)
            logger.info("Load command line options from {}.", file_path)
            return data
        except:  # noqa
            return {}


def generate_all_settings(
    settings_models: Union[None, str, Iterable[str]] = None,
    name: Optional[str] = None,
    mixins: Optional[List[type]] = None,
) -> Type[T]:
    if not name:
        name = "UniversalSettings"
    if mixins is None:
        mixins = []
    base_classes = parse_settings_models(settings_models)
    new_class = type(
        name,
        (
            *mixins,
            *base_classes,
            PydanticBaseSettings,
        ),
        {},
    )
    return new_class  # type: ignore


class SettingsProxy(Generic[T]):
    def __init__(self) -> None:
        self._settings: Optional[T] = None

    def __getattr__(self, attr: str) -> Any:
        if attr in self.__dict__:
            return self.__dict__[attr]
        if self._settings is None:
            raise ValueError("settings not initialized")
        return getattr(self._settings, attr)

    def __str__(self) -> str:
        return str(self._settings)

    def __repr__(self):
        return repr(self._settings)


def init_settings(
    settings_cls: Type[T], overwrite: bool = True
) -> Union[T, SettingsProxy[T]]:
    _settings = getattr(settings, "_settings", None)
    if _settings is None or overwrite:
        _settings = settings_cls()
        setattr(settings, "_settings", _settings)
    return _settings


settings: Union[BaseSettings, SettingsProxy[Any]] = SettingsProxy()


def get_settings_proxy():
    return settings



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydantic_universal_settings']

package_data = \
{'': ['*']}

install_requires = \
['click-option-group>=0.5.3,<0.6.0',
 'psutil>=5.8.0,<6.0.0',
 'pydantic[dotenv]>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'pydantic-universal-settings',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'tc-imba',
    'author_email': 'liuyh970615@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import asyncio
from functools import lru_cache, wraps
from typing import Any, Callable, Iterable, List, Optional, Tuple, Type, Union

import click
from click_option_group import optgroup
from pydantic import BaseModel

from pydantic_universal_settings.main import parse_settings_models


def parse_settings_group(cls: Type[BaseModel]) -> Tuple[str, str]:
    docstring = cls.__doc__ or ""
    help_index = docstring.find("\n\n")
    if help_index < 0:
        if docstring:
            name = docstring
        else:
            name = cls.__name__
        help_str = ""
    else:
        name = docstring[:help_index]
        help_str = docstring[help_index + 2 :]
    return name, help_str


def get_option_name(name: str, _type: type) -> str:
    name = name.replace("_", "-")
    if _type == bool:
        return f"--{name}/--no-{name}"
    return f"--{name}"


def get_option_help(description: Optional[str], default: Any) -> str:
    result = (description or "") + "\n"
    if default is not None:
        result += f"[default: {str(default)}]"
    return result


@lru_cache()
def get_settings_class_options(settings_class: Type[BaseModel]) -> List[Any]:
    options = []
    name, help_str = parse_settings_group(settings_class)
    options.append(optgroup.group(name=name, help=help_str))
    option_base = optgroup

    for key, value in settings_class.__fields__.items():
        opt_name = get_option_name(key, value.type_)
        help_str = get_option_help(value.field_info.description, value.default)
        options.append(
            option_base.option(
                opt_name,
                type=value.type_,
                help=help_str,
                default=None,
            )
        )
    return options


def get_global_options(settings_models: List[Type[BaseModel]]) -> List[Any]:
    global_options = [click.argument("args", nargs=-1)]
    for settings_class in settings_models:
        global_options.extend(get_settings_class_options(settings_class))
    return global_options


def add_options(options: List[Any]) -> Callable[..., Callable[..., Any]]:
    def _add_options(func: Any) -> Callable[..., Any]:
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def command_start(
    settings_models: Union[None, str, Iterable[str]] = None,
    name: str = None,
    group: Optional[click.Group] = None,
) -> Any:
    parsed_settings_models = parse_settings_models(settings_models)
    context_settings = dict(
        help_option_names=["-h", "--help"],
        max_content_width=88,
    )
    if group is None:
        command_decorator = click.command
    else:
        command_decorator = group.command

    def decorator(func: Any) -> Any:
        @command_decorator(name=name, context_settings=context_settings)
        @add_options(get_global_options(parsed_settings_models))
        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs, __settings_models=parsed_settings_models)

        return wrapped

    return decorator


cli_settings = {}


def command_end() -> Any:
    def decorator(func: Any) -> Any:
        @wraps(func)
        def wrapped(args: Any, **kwargs: Any) -> Any:
            if "__settings_models" not in kwargs:
                raise SystemError(
                    "Please use cli_command_start before cli_command_end!"
                )
            settings_models = kwargs.pop("__settings_models")
            fields = set()
            for settings_class in settings_models:
                fields.update(settings_class.__fields__.keys())

            new_kwargs = {}
            cli_settings["__not_empty"] = True
            for key, value in kwargs.items():
                if key in fields:
                    if value is not None:
                        cli_settings[key] = value
                else:
                    new_kwargs[key] = value
            return func(*args, **new_kwargs)

        return wrapped

    return decorator


def command(
    settings_models: Union[None, str, Iterable[str]] = None,
    name: str = None,
    group: Optional[click.Group] = None,
) -> Any:
    def decorator(func: Any) -> Any:
        @command_start(settings_models=settings_models, name=name, group=group)
        @command_end()
        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        return wrapped

    return decorator


def async_command(f: Any) -> Any:
    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(f(*args, **kwargs))

    return wrapper
